#include "Behavior.h"

#include <QtDebug>

#include "Sensor.h"
#include "VictimRFIDSensor.h"
#include "Agent.h"

BehaviorController::BehaviorController(MotionController* motionController, Manifold* map) {
    this->motionController = motionController;
	this->map = map;

	this->behaviors.insert(WAITFORLASER,    new BehaviorWaitForLaser(this));
    this->behaviors.insert(EXPLORE,         new BehaviorExplore(this));
    this->behaviors.insert(GETVICTIMID,     new BehaviorGetVictimID(this));
    this->behaviors.insert(GETVICTIMSTATUS, new BehaviorGetVictimStatus(this));

    this->changeTo(EXPLORE);

}
BehaviorController::~BehaviorController() {
    QMapIterator< BehaviorType, Behavior*> iterator(this->behaviors);
    while (iterator.hasNext()) {
        iterator.next();
        delete iterator.value();
    }
    this->behaviors.clear();
}

void BehaviorController::notifyUpdate(Sensor* sensor) {
    if (this->currentBehavior) {
        this->currentBehavior->processUpdate(sensor);
    }        
}
void BehaviorController::changeTo(BehaviorType behavior) {
    if (this->behaviors.contains(behavior)) {
		previousBehaviorType = currentBehaviorType;
		currentBehaviorType = behavior;
        this->currentBehavior = this->behaviors[behavior];
        qDebug() << QString("BEHAVIOR CHANGED TO: %1 (%2)").arg(this->currentBehavior->getName()).arg(this->currentBehavior->getDescription());
    }
}
void BehaviorController::executeMotion(MotionType motion) {
    this->motionController->execute(motion);
}


void BehaviorWaitForLaser::processUpdate(Sensor* sensor)
{
	if (sensor->getType() == LaserRangeSensor::LASERRANGE) {
		this->controller->changeToPrevious();
	}
}

void BehaviorExplore::processUpdate(Sensor* sensor) {
	if (sensor->getType() == RFIDSensor::RFID) {
		QMap<QString, RFIDData> rfid = ((RFIDSensor*) sensor)->getData();
		
		QList< RFIDData > values = rfid.values();
        for(int i=0; i<values.size(); i++) {
        		if( !this->controller->getMap()->hasRFID( values[i].ID ) ) {
        			this->controller->changeTo(WAITFORLASER);
        			return;
        		}
        }
	} else if (sensor->getType() == VictimRFIDSensor::VICTIMRFID) {
        QMap<QString, VictimRFIDData> rfid = ((VictimRFIDSensor*) sensor)->getData();
        
        IdentificationLevel maxLevel = CANNOTSEE;
        
        QList< VictimRFIDData > values = rfid.values();
        for(int i=0; i<values.size(); i++) {
        		if( maxLevel < values[i].level )
        			maxLevel = values[i].level;
        			
        		if( maxLevel > CANSEE && !this->controller->getMap()->hasVictim( values[i].ID ) ) {
        			this->controller->changeTo(WAITFORLASER);
        			return;
        		}
        }
        
        qDebug() << "explore: got victim " << values[0].dx << values[0].dy << maxLevel;
        
        switch (maxLevel) {
            case CANSEE :
                this->controller->changeTo(GETVICTIMID);
                break;
            case CANSEEVICTIM :
                this->controller->changeTo(GETVICTIMSTATUS);
                break;
            default :
                break;
        }
    }
	this->controller->executeMotion(JUSTGO);
}




void BehaviorGetVictimID::processUpdate(Sensor* sensor) {
	if(sensor->getType() == LaserRangeSensor::LASERRANGE) {
		counter ++;
	}
    else if (sensor->getType() == RFIDSensor::RFID) {
		this->controller->changeTo(WAITFORLASER);
	} else if (sensor->getType() == VictimRFIDSensor::VICTIMRFID) {
        QMap<QString, VictimRFIDData> rfid = ((VictimRFIDSensor*) sensor)->getData();
        
        // find first unknown one
        QString vic;
        QList<QString> keys = rfid.keys();
        for(int i=0; i<keys.size(); i++)
        {
        		if( rfid[keys[i]].level == CANSEE || !controller->getMap()->hasVictim(keys[i]) )
        		{
        			vic = keys[i];
        			break;
        		}
        }
        
        if( vic.length() == 0 )
        {
        		qDebug() << "all sensed victims known already";
        }
        
        switch (rfid[vic].level) {
            case CANSEEFALSEALARM :
                this->controller->changeTo(EXPLORE);
                break;
            case CANSEEVICTIM :
                this->controller->changeTo(GETVICTIMSTATUS);
                break;
            default :
                break;
        }
        
        this->controller->getMotionController()->attractionPoint.x = rfid[vic].dx;
        this->controller->getMotionController()->attractionPoint.y = rfid[vic].dy;
        this->controller->executeMotion(GOTOVICTIM);
        
        qDebug() << "current victim at: " << rfid[vic].dx << rfid[vic].dy;
        
        counter = 0;
    }
    
    if( counter > 5 ) // we haven't gotten interesting messages for 5 laserscans
    {
    		counter = 0;
    		this->controller->changeTo(EXPLORE);
    }
}




void BehaviorGetVictimStatus::processUpdate(Sensor* sensor) {
	
	if(sensor->getType() == LaserRangeSensor::LASERRANGE) {
		counter ++;
	}
    else if (sensor->getType() == RFIDSensor::RFID) {
		this->controller->changeTo(WAITFORLASER);
	} else if (sensor->getType() == VictimRFIDSensor::VICTIMRFID) {
		QMap<QString, VictimRFIDData> rfid = ((VictimRFIDSensor*) sensor)->getData();
		
		QString vic;
        QList<QString> keys = rfid.keys();
        for(int i=0; i<keys.size(); i++)
        {
        		if( controller->getMap()->getVictimStatus(keys[i]).length()!=0 )
        		{
        			vic = keys[i];
        			break;
        		}
        		
        		qDebug() << "victim status: " << keys[i] << controller->getMap()->getVictimStatus(keys[i]);
        }
        
        if( vic.length() == 0 )
        {
        		qDebug() << "all sensed victims known already";
        		this->controller->changeTo(EXPLORE);
        		return;
        }
		
        
        switch (rfid[vic].level) {
            case CANSEEVICTIMSTATUS :
                this->controller->changeTo(EXPLORE);
                break;
            default :
                break;
        }
        
        this->controller->getMotionController()->attractionPoint.x = rfid[vic].dx;
        this->controller->getMotionController()->attractionPoint.y = rfid[vic].dy;
        this->controller->executeMotion(GOTOVICTIM);
        
        qDebug() << "(status) current victim at: " << rfid[vic].dx << rfid[vic].dy;
        
        counter = 0;
    }
    
    if( counter > 5 ) // we haven't gotten interesting messages for 5 laserscans
    {
    		counter = 0;
    		this->controller->changeTo(EXPLORE);
    }
}

