#include "Motion.h"
#include "ScanUtils.h"

#include <math.h>

using namespace std;


MotionController::MotionController(PositionActuator* actuator) {
    this->actuator = actuator;
    this->motions.insert(JUSTGO, new MotionExplore(this));
    this->motions.insert(GOTOVICTIM, new MotionAttractToVictim(this));
    this->currentMotionType = NONE;
    this->currentMotion     = 0;
}
MotionController::~MotionController() {
    QMapIterator< MotionType, Motion*> iterator(this->motions);
    while (iterator.hasNext()) {
        iterator.next();
        delete iterator.value();
    }
    this->motions.clear();
}
void MotionController::notifyUpdate(Sensor* sensor) {
    if (this->currentMotion) {
        this->currentMotion->processUpdate(sensor);
    }        
}
void MotionController::execute(MotionType motion) {

    // do not disturb an ongoing motion if not necessary
    if (motion == this->currentMotionType)
        return;

    if (this->currentMotion) {
        this->currentMotion->deactivate();
    }
    if (this->motions.contains(motion)) {
        this->currentMotionType = motion;
        this->currentMotion     = this->motions[motion];
        this->currentMotion->activate();
    } else {
        this->currentMotionType = NONE;
        this->currentMotion     = 0;
    }
}
void MotionController::move(float translation, float rotation) {
    this->actuator->drive(translation, rotation);
}



void Motion::processUpdate(Sensor* sensor) {
    QString type = sensor->getType();
    if (type == INUSensor::INU) {
        this->process(((INUSensor*) sensor)->getData());
    } else if (type == LaserRangeSensor::LASERRANGE) {
        this->process(((LaserRangeSensor*) sensor)->getNewestData());
    } else if (type == OdometrySensor::ODOMETRY) {
        this->process(((OdometrySensor*) sensor)->getData());
    } else if (type == RFIDSensor::RFID) {
        this->process(((RFIDSensor*) sensor)->getData());
    } else if (type == SonarSensor::SONAR) {
        this->process(((SonarSensor*) sensor)->getData());
    } else if (type == TouchSensor::TOUCH) {
        this->process(((TouchSensor*) sensor)->getData());
    } else if (type == VictimRFIDSensor::VICTIMRFID) {
        this->process(((VictimRFIDSensor*) sensor)->getData());
    } 
}

void MotionAttractToVictim::process(const LaserData& range)
{
	double avgWest  = 0.0;
    for (uint i = 0; i < range.length/3; i++) {
        avgWest += (1.0/range.length*3) * range.data[i];
    }
    double minNorth = range.data[range.length/2];
    for (uint i = range.length/5; i < range.length*4/5; i++) {
        minNorth = min(minNorth, range.data[i]);
    }
    double avgEast = 0.0;
    for (uint i = range.length*2/3; i < range.length; i++) {
        avgEast += (1.0/range.length*3) * range.data[i];
    }
    
    if( minNorth < .55 )
    {
    		avoiding = true;
    		Point2 victim = controller->attractionPoint;
    		
    		if( victim.y/victim.x > .18 )
			this->controller->move(0, -0.8f);
		else if(victim.y/victim.x < -.18)
			this->controller->move(0, 0.8f);
    } else {
    		avoiding = false;
    }
}

void MotionAttractToVictim::process(const QMap<QString, VictimRFIDData> & )
{
	Point2 victim = controller->attractionPoint;
	
	if(!avoiding)
	{
		if( victim.y/victim.x > .25 )
			this->controller->move(0, -0.7f);
		else if(victim.y/victim.x < -.25)
			this->controller->move(0, 0.7f);
		else
			this->controller->move(.4f * .5f*victim.x, 0);
	}
}

MotionExplore::MotionExplore(MotionController* controller) 
: Motion(controller) {
    this->isTurningEast = false;
    this->isTurningWest = false;
    this->isRetreating  = false;
    this->justRetreated = false;
    this->turningAfterRetreated = false;
    this->freeSpace = 0;
}

void MotionExplore::process(const SonarData& sonar)
{
	//qDebug() << "sonar F: " << sonar.F1 << sonar.F2 << sonar.F3 << sonar.F4 << sonar.F5 << sonar.F6 << sonar.F7 << sonar.F8;
	//qDebug() << "sonar R: " << sonar.R1 << sonar.R2 << sonar.R3 << sonar.R4 << sonar.R5 << sonar.R6 << sonar.R7 << sonar.R8;
	
	double minF = sonar.F2;
	if(sonar.F4 < minF)
		minF = sonar.F4;
	if(sonar.F5 < minF)
		minF = sonar.F5;
	if(sonar.F7 < minF)
		minF = sonar.F7;
		
	double minR = sonar.R2;
	if(sonar.R4 < minR)
		minR = sonar.R4;
	if(sonar.R5 < minR)
		minR = sonar.R5;
	if(sonar.R7 < minR)
		minR = sonar.R7;
		
	if( isRetreating && ( minR < .2 || sonar.R3 > .6 || sonar.R6 > .6 || sonar.R3 < .15 || sonar.R6 < .15 ) )
	{
		// stop retreating and turn
		isRetreating = false;
		justRetreated = true;
		
		//qDebug() << "******* SONAR stopped retreat";
	}
	
	if( !isRetreating && ( minF < .2 || sonar.F3 > .6 || sonar.F6 > .6 || sonar.F3 < .15 || sonar.F6 < .15 ) )
	{
		isRetreating = true;
		startRetreat = clock();
		
		qDebug() << "******* SONAR initiated retreat";
	}
	
	this->move();
}

void MotionExplore::process(const LaserData& range) {
    double avgWest  = 0.0;
    for (uint i = 0; i < range.length/3; i++) {
        avgWest += (1.0/range.length*3) * range.data[i];
    }
    double minNorth = range.data[range.length/2];
    for (uint i = range.length/5; i < range.length*4/5; i++) {
        minNorth = min(minNorth, range.data[i]);
    }
    double avgEast = 0.0;
    for (uint i = range.length*2/3; i < range.length; i++) {
        avgEast += (1.0/range.length*3) * range.data[i];
    }

    //qDebug() << QString("W/N/E = (%1) / (%2) / (%3)").arg(avgWest).arg(minNorth).arg(avgEast);

    this->canGoNorth = (minNorth > .55);
    if (avgEast > .3 && avgWest > .3) {
        this->eastLooksPromising = (avgEast > avgWest);
        this->westLooksPromising = (avgWest > avgEast);
    } else {
        this->eastLooksPromising = (avgEast > .3);
        this->westLooksPromising = (avgWest > .3);
    }
    
    
    freeSpace = minNorth / range.maxRange * 5; // scale 1 = 5 meters?
    if( freeSpace > 1 )
    		freeSpace = 1;

    this->move();
}

void MotionExplore::process(const INUData& inu) {

    // yaw   is around z-axis, this may have any value
    // pitch is around y-axis, danger of falling backwards
    // roll  is around x-axis, danger of falling sideways

	if( isRetreating && ( (clock() - startRetreat)/CLOCKS_PER_SEC < ( Random::generate(2.0,4.0)) ) )
	{
		this->move();
		return;
	}
	
	bool wasRetreating = isRetreating;
	
    bool somethingWrong = (fabs(inu.dpitch) > .45)
                            || (fabs(inu.droll) > .45);

	if( somethingWrong && isRetreating )
	{
		isRetreating = false;
	}
	else if(somethingWrong && !isRetreating )
	{
		isRetreating = true;
		
		qDebug() << "****** INU initiated retreat " << inu.dpitch << inu.droll;
	}
	else if(!somethingWrong)
	{
		isRetreating = false;
	}

	if( isRetreating )
		startRetreat = clock();
		
	if( !isRetreating && wasRetreating )
	{
		justRetreated = true;
	}

    this->move();
}

void MotionExplore::deactivate()
{
	this->controller->move(0, 0);
}

void MotionExplore::move() {

	if( turningAfterRetreated && ( ((float)(clock() - startJustRetreated))/CLOCKS_PER_SEC < ( Random::generate(1.0,3.5)) ) ) {
		return;
	} else {
		turningAfterRetreated = false;
	}

    if (isRetreating) {
        // retreat
        //qDebug() << "Retreat";
        this->isTurningEast = false;
        this->isTurningWest = false;
    } else if (justRetreated) {
    		// turn randomly for 1 sec
    		isTurningEast = Random::flipCoin();
    		isTurningWest = !isTurningEast;
    		startJustRetreated = clock();
    		turningAfterRetreated = true;
    		justRetreated = false;
    		
	} else if (this->canGoNorth) {
        // go north
        //qDebug() << "North";
        this->isTurningEast = false;
        this->isTurningWest = false;
    } else {
        // continue movement that was already started
        if (this->isTurningEast) {
            // no changes
        } else if (this->isTurningWest) {
            // no changes
        } else {
            // start a turn
            if (this->eastLooksPromising == this->westLooksPromising) {
                // it does not really matter what we do ...
                if (Random::flipCoin()) {
                    //qDebug() << "East after flipping a coin";
                    this->isTurningEast = true;
                } else { 
                    //qDebug() << "West after flipping a coin";
                    this->isTurningWest = true;
                }
            } else if (this->eastLooksPromising) {
                //qDebug() << "East";
                this->isTurningEast = true;
            } else if (this->westLooksPromising) {
                //qDebug() << "West";
                this->isTurningWest = true;
            }
        }
    }

    if (this->isRetreating) {
        this->controller->move(-.5f, 0);
    } else if (this->isTurningEast) {
        this->controller->move(0, -2.5f);
    } else if (this->isTurningWest) {
        this->controller->move(0, 2.5f);
    } else {
        this->controller->move(0.4f + freeSpace*1.0f, 0);
    }

}

