#include "Robot.h"

// no forwarding required anymore
#include "RobotRefDevice.h"

#include <iostream>
#include <QRegExp>
#include <QtDebug>

const int Robot::TIMEOUT = 5 * 1000; // 5 seconds
const int Robot::BUFSIZE = 100 * 1024; // chars , 100k

Robot::Robot(const QString &host, const quint16 port) {

    this->host = host;
	this->port = port;

    this->uniqueID = QUuid::createUuid();
    
    this->botName = "Jimmy";
	this->botModel = "USARBot.P2AT";
	this->botPos = "0,0,0";
	this->botRot = "0,0,0";
}

void Robot::connect() {
	server.connectToHost(host,port);

	if(!server.waitForConnected(TIMEOUT)) {
        throw Exception("Couldn't connect to " + host + "! " + server.errorString());
	}

	spawnRobot();
	
}

void Robot::stop() {
	keepRunning = false;
}


QUuid Robot::getUniqueID() {
    return this->uniqueID;
}

QString Robot::getName() {
    return this->botName;
}

QString Robot::getUniqueName() {
    return this->botName + "_" + this->uniqueID.toString();
}

void Robot::setName(QString botName) {
	this->botName = botName;
}

void Robot::setModel(QString botModel) {
	this->botModel = botModel;
}

void Robot::setPos(QString botPos) {
	this->botPos = botPos;
}

void Robot::setRot(QString botRot) {
	this->botRot = botRot;
}

void Robot::disconnect() {
	stop();
	server.flush();
	server.close();
}

void Robot::putCommand(QString cmd) {
	QMutexLocker locker(&outputMutex);
	output.enqueue(cmd);
}

void Robot::addDevice(Device *device) {
	
	QString type = device->getType();
	QString name = device->getName();
	
	qDebug() << "Adding Device: " << type << " - " << name;

	if(!devices.contains(type)) {
		QMap< QString, Device* > newmap;
		newmap.insert(name, device);

		devices.insert(type,newmap);
	} else {
		QMap< QString, Device* > &map = devices[type];

		map.insert(name,device);
	}
	
	RobotRefDevice * ref = dynamic_cast<RobotRefDevice*>(device);
	if(ref!=NULL) {
		qDebug() << "Device needs a ref, setting";
		ref->setRobot(this);	
	}
}


void Robot::run() {

	keepRunning = true;

	char* buffer = new char[BUFSIZE];
	buffer[0] = 0;
	int index = 0;
	bool done = false;
	
	QRegExp start("^(\\S+)\\s*");
	QRegExp bits("\\s*(\\{([^\\} ]+) ([^\\} ]+)\\}|\\{([^\\} ]+) ([^\\} ]+) ([^\\} ]+) ([^\\}]+)\\})\\s*");
	
    while(keepRunning) {

		if(keepRunning) {
			QMutexLocker locker(&outputMutex);
			
			QString str;
			while(!output.isEmpty()) {
				str = output.dequeue();
				//qDebug() << "SENDING: " << str;
				
				server.write(str.toAscii());
			}
		}

		index = 0;
		buffer[0]=0;
		done = false;
		
		QMap< QString, QString > dataMap;

		while( !done && keepRunning ) {
			if( server.state() != QAbstractSocket::ConnectedState ) {
				keepRunning = false;
				qDebug() << "Socket closed?!? reason: " << server.errorString();
				break;
			}
			if(!server.waitForReadyRead(TIMEOUT)) {
				//throw Exception("Couldn't read from socket: "+server.errorString());
                continue;
			}
			if( server.canReadLine() ) {
				index = server.readLine(buffer,BUFSIZE);
				buffer[index-2] = 0;
				done = true;
			}
		}
		
		// make sure this all gets cleaned up each cycle
		if(keepRunning) {
			
			QString msg(buffer);
	
			if(index>0) {
				
                //qDebug() << "RECV: " << buffer;
	
				// parse!
				
				dataMap.clear();
				
				if( start.indexIn(msg) > -1) {
					
					dataMap.insert("msg", start.cap(1));
					
					int pos = start.matchedLength();
					
					while( (pos = bits.indexIn(msg, pos)) != -1 ) {
						
						if( bits.cap(2).length() != 0 ) {
							dataMap.insert( bits.cap(2), bits.cap(3) );
						} else if( bits.cap(4).length() != 0 ) {
							dataMap.insert( bits.cap(5), bits.cap(7) );
						}
						
						pos += bits.matchedLength();
					}
					
					if(! distribute(&dataMap) ) {
						//qDebug() << "Message was: " << msg;
					}
					
				} else {
					qDebug() << "Not valid message!:" << msg;
					continue;
				}
			}
		}
	}

}

bool Robot::distribute(QMap<QString, QString> *map) {

	QString msg = (*map)["msg"];
	QString type = (*map)["Type"];
	QString name(Device::ANY);
	if(map->contains("Name")) {
		name = (*map)["Name"];
	}
	
	bool distributed = false;

    // check for devices that process ANY type
    if( devices.contains(Device::ANY) ) {
		QMap<QString, Device* > &devs = devices[Device::ANY];
		
		if( name!=Device::ANY && devs.contains(name) ) {
			Device *dev = devs[name];
			if(dev->acceptsMsg(msg)) {
				dev->process(map);
				distributed = true;
			}
		} else if( devs.contains(Device::FALLBACK) ) {
			Device *dev = devs[Device::FALLBACK];
			if(dev->acceptsMsg(msg)) {
				dev->process(map);
				distributed = true;
			}
		}
		
		if( devs.contains(Device::ANY) ) {
			Device *dev = devs[Device::ANY];
			if(dev->acceptsMsg(msg)) {
				dev->process(map);
				distributed = true;
			}
		}
	}

    // do the specific device type
	if( devices.contains(type) ) {
		QMap<QString, Device* > &devs = devices[type];
		
		if( name!=Device::ANY && devs.contains(name) ) {
			Device *dev = devs[name];
			if(dev->acceptsMsg(msg)) {
				dev->process(map);
				distributed = true;
			}
		} else if( devs.contains(Device::FALLBACK) ) {
			Device *dev = devs[Device::FALLBACK];
			if(dev->acceptsMsg(msg)) {
				dev->process(map);
				distributed = true;
			}
		}
		
		if( devs.contains(Device::ANY) ) {
			Device *dev = devs[Device::ANY];
			if(dev->acceptsMsg(msg)) {
				dev->process(map);
				distributed = true;
			}
		}
	}
	
	if(!distributed) {
		qDebug() << "message not processed: " << msg << " " << type;
	}
	
	return distributed;
}

void Robot::spawnRobot() {
	putCommand(
		QString("INIT {ClassName %1} {Name %2} {Location %3} {Rotation %4}\r\n")
			.arg(botModel)
			.arg(botName)
			.arg(botPos)
			.arg(botRot)
		);
}
