

#include "UsarAgent.h"

#include <QtDebug>
#include <QStringList>


UsarAgent::UsarAgent() : Agent(), detector(30,15,.05)
{
	usarclient = NULL;
	this->renderingmode = Agent::RENDERINGMODE_SIMPLE;
}

UsarAgent::~UsarAgent() {
}

void UsarAgent::configure(QString model, QString name, QString position, QString rotation) {
    Agent::configure(model, name, position, rotation);

	this->renderingmode = UsarAgent::RENDERINGMODE_SIMPLE;

    QStringList list = position.split(",");
    this->initialPose.x = list[0].toDouble();
    this->initialPose.y = list[1].toDouble();

    list = rotation.split(",");
    this->initialPose.rot = list[2].toDouble();

    this->position   = new PositionActuator(this);
    this->keyboard   = new KeyboardSensor();
    this->inu        = new INUSensor("INU");
    this->odometry   = new OdometrySensor("Odometry");
    this->laser      = new LaserRangeSensor("Scanner1");
    this->rfid       = new RFIDSensor("RFID");
    this->sonar      = new SonarSensor();
    this->touch      = new TouchSensor();
    this->victimrfid = new VictimRFIDSensor("VictRFID");

    this->mount(keyboard);
    this->mount(inu);
    this->mount(odometry);
    this->mount(laser);
    this->mount(rfid);
    this->mount(sonar);
    this->mount(touch);
    this->mount(victimrfid);

    this->manifold = new Manifold();
    this->matcher  = new WeightedScanMatcher();
    this->slam     = new ManifoldSlam(matcher, manifold);

    this->motionController   = new MotionController(this->position);
    this->behaviorController = new BehaviorController(this->motionController, manifold);

    this->keyboard->start();

}

void UsarAgent::connectToMapServer(QString host, uint port)
{
	if( usarclient == NULL )
	{
		usarclient = new UsarClient(host, port);
		usarclient->connect();
		
		slam->addListener(usarclient);
	}
	else
	{
		qDebug() << "UsarAgent already connected with a UsarClient!";
	}
}

void UsarAgent::notifyUpdate(Sensor* sensor) {
    Agent::notifyUpdate(sensor);

    // first notify behavior control, it may want to change the current motion
    this->behaviorController->notifyUpdate(sensor);
    this->motionController->notifyUpdate(sensor);

    // forward relevant sensor reading to SLAM/Manifold
	if (sensor->getType() == LaserRangeSensor::LASERRANGE) {
        this->observed(((LaserRangeSensor*) sensor)->getNewestData());
    } else if (sensor->getType() == VictimRFIDSensor::VICTIMRFID) {
        this->observed(((VictimRFIDSensor*) sensor)->getData());
    } else if (sensor->getType() == RFIDSensor::RFID) {
        this->observed(((RFIDSensor*) sensor)->getData());
    }
   
}

void UsarAgent::observed(LaserData range) {
    qDebug() << QString("RANGE SCAN RECIEVED");

	{
		INUData      inu   = this->inu->getData();
		if( fabs(inu.dpitch) > M_PI/4.0 || fabs(inu.droll) > M_PI/4.0 ) {
			this->keepRunning = false;
			return;
		}

		bool isNotOK = true;
		for(int i=0; i<range.length; i++)
		{
			isNotOK = isNotOK && ( range.data[i] <= range.minRange);
		}
		if (isNotOK) {
			this->keepRunning = false;
			return;
		}
	}

	ScanObservation* patch = this->slam->getCurrentPatch();
    Pose2D origin;
    if (patch) {
        //origin   = patch->pose;
		origin   = this->slam->getCurrentPose();
        origin.x = origin.x / 1000;
        origin.y = origin.y / 1000;
    } else {
        origin = this->initialPose;
    }
    
    /*
    qDebug() << "range.length="<<range.length<<";";
    qDebug() << "range.fov="<<range.fov<<";";
    qDebug() << "range.resolution="<<range.resolution<<";";
    qDebug() << "range.maxRange="<<range.maxRange<<";";
    qDebug() << "range.minRange="<<range.minRange<<";";
    qDebug() << "range.data = new int["<<range.length<<"];";
	for(int i=0; i<range.length; i++)
	{
		qDebug() << "range.data["<<i<<"]="<<range.data[i]<<";";
	}
	*/


	/*
    OdometryData dpose = this->odometry->getData();
	//INUData      inu   = this->inu->getData();

    // the  dpose is wrt the current patch
    // so here it is converted to map (global) coords
    Matrix3 rotmx;
	rotmx.fromPose2D(origin);
    Vector2 vdpose(dpose.dx, dpose.dy);
    //rotmx.setRotation(-pose.rot);
    vdpose = rotmx.multiply(vdpose);

    origin.x   = vdpose.x;
    origin.y   = vdpose.y;
	origin.rot = origin.rot + dpose.dtheta;
	//*/

    ScanObservation* scan = new ScanObservation(1000, origin.x, origin.y, origin.rot, range);
    
    if( !previousRFIDs.isEmpty() ) {
    		QList<RFIDData> rfids = previousRFIDs.values();
		for(int i=0; i<rfids.size(); i++)
		{
			qDebug() << QString("I SAW RFID %1").arg(rfids[i].ID);
			scan->addRFID(rfids[i]);
		}
		previousRFIDs.clear();
    }
    
    if( !previousVictimRFIDs.isEmpty() ) {
    		QList<VictimRFIDData> rfids = previousVictimRFIDs.values();
		for(int i=0; i<rfids.size(); i++)
		{
			if (rfids[i].level >= CANSEEVICTIM && rfids[i].ID != "FalsePositive") {
				qDebug() << QString("I SAW VICTIM %1 (Status %2) (loc %3,%4,%5)")
					.arg(rfids[i].ID)
					.arg(rfids[i].status)
					.arg(rfids[i].dx)
					.arg(rfids[i].dy)
					.arg(rfids[i].dz);
				scan->addVictim(rfids[i]);
			}
		}
		previousVictimRFIDs.clear();
    }
    
    this->slam->processScan(scan);
    this->odometry->reset();
    //this->inu->reset();
    
    
    // *********************** BUGGY hough transform **********************
    /*
    if( false )
    {
	    QList<Point2> potentialVictims = detector.getCircles(range);
	    if(potentialVictims.size() > 0)
	    {
	    		Point2 vic = potentialVictims[0];
	    		VictimRFIDData data;
	    		data.dx = vic.x;
	    		data.dy = vic.y;
	    		data.dz = -.85;
	    		
	    		data.ID = "None";
	    		data.status = "None";
	    		data.level = CANSEE;
	    		
	    		qDebug() << QString("injected scanner-based VICTIM (loc %3,%4,?)").arg(data.dx).arg(data.dy);
			
			victimrfid->injectObservation(data);
	    }
    }
    //*/
    // *********************************************************************
}

void UsarAgent::observed(QMap<QString, RFIDData> rfid) {
	qDebug() << "************ new RFID data";
	previousRFIDs = rfid;
}

void UsarAgent::observed(QMap<QString, VictimRFIDData> rfid) {
	qDebug() << "************ new VictimRFID data";
	QList<VictimRFIDData> rfids = rfid.values();
	for(int i=0; i<rfids.size(); i++)
	{
		qDebug() << QString("I SAW VICTIM %1 (Status %2)").arg(rfids[i].ID).arg(rfids[i].status);
		this->slam->processVictim(rfids[i]);
	}
	
	previousVictimRFIDs = rfid;
}




void UsarAgent::saveMap(int pxPerMeter, QString filename) {

    if (!filename.toLower().endsWith(".png"))
        filename.append(".png");

	Manifold* subset = this->manifold->getConnectedSubSet();
	Bounds bounds    = subset->computeBounds();
	//Bounds bounds    = this->manifold->computeBounds();

    qDebug() << "SAVING MAP ...";
    qDebug() << QString("-- Map bounds are (%1,%2 %3,%4)").arg(bounds.top).arg(bounds.left).arg(bounds.height).arg(bounds.width);
    qDebug() << QString("-- Full Manifold has %1 nodes and %2 links").arg(this->manifold->nodeCount()).arg(this->manifold->linkCount());
    qDebug() << QString("-- Connected subset has %1 nodes and %2 links").arg(subset->nodeCount()).arg(subset->linkCount());

	int iHeight = (float)pxPerMeter * bounds.height / 1000;
	int iWidth  = (float)pxPerMeter * bounds.width / 1000;

	qDebug() << QString("-- Making image of size (%1,%2)").arg(iWidth).arg(iHeight);

	// keep aspect ratio of bounds
    QImage img = QImage(iWidth, iHeight, QImage::Format_ARGB32_Premultiplied);
    
    // make sure maprenderer gets destroyed so the painter gets destroyed too
    {
		ManifoldRenderer* renderer = NULL;
		switch (this->renderingmode) {
			case Agent::RENDERINGMODE_NONE:
				break;
			case Agent::RENDERINGMODE_SIMPLE:
				renderer = new QtSimpleManifoldRenderer(&img, bounds);
				break;
			case Agent::RENDERINGMODE_SLEEK:
				renderer = new QtManifoldRenderer(&img, bounds);
				break;
		}
		if (renderer != NULL) {
			//this->manifold->render(renderer);
			subset->render(renderer);
			delete renderer;
			renderer = NULL;
		}
    }

    if (img.save(filename, "PNG")) {
        qDebug() << QString("-- Map saved to %1").arg(filename);
        qDebug() << QString("-- Done");
    } else {
        qDebug() << QString("-- Could not save map to %1").arg(filename);
    }

	delete subset;

}


void UsarAgent::saveRFIDs(QString filename) {
	QFile file(filename);
	if (file.open(QIODevice::WriteOnly | QFile::Truncate)) {

        qDebug() << "Saving RFIDs to file: " << filename;

		// shoud be doing breadth first
		
        QTextStream stream(&file);
        
        
        QList<RFIDData> rfids;
        {
        		GraphReadLocker lock(manifold);
       		rfids = this->manifold->getRFIDs();
        }
        
        for(int i=0; i<rfids.size(); i++)
        {
    		qDebug() << "- " << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0;
    		stream << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0 << endl;
        }
        
        /*
        QMapIterator< QString, RFIDObservation* > rfidIterator = this->manifold->rfidIterator();
        RFIDObservation* rfid;
        RFIDData data;
        while (rfidIterator.hasNext()) {
            rfidIterator.next();
            rfid = rfidIterator.value();
            data = rfid->getRFID();
            stream << data.ID << "," << data.dx  << "," << data.dy  << "," << data.dz << endl;
        }
        //*/

        stream << "END" << endl;
	}	
}

void UsarAgent::saveVictims(QString filename) {

	QFile file(filename);
	if (file.open(QIODevice::WriteOnly | QFile::Truncate)) {
		
        qDebug() << "Saving Victims to file: " << filename;

		
	    QTextStream stream(&file);

		QList<VictimRFIDData> rfids;
		{
			GraphReadLocker lock(manifold);
			rfids = this->manifold->getVictimRFIDs();
		}
        for(int i=0; i<rfids.size(); i++)
        {
        		qDebug() << "- " << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0 << "," << this->manifold->getVictimStatus( rfids[i].ID );
        		stream << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0 << "," << this->manifold->getVictimStatus( rfids[i].ID ) << endl;
        }
		//*/

		/*
        QMapIterator< QString, VictimObservation* > victimIterator = this->manifold->victimIterator();
        VictimObservation* victim;
        VictimRFIDData data;
        while (victimIterator.hasNext()) {
            victimIterator.next();
            victim = victimIterator.value();
            data   = victim->getRFID();
            stream << data.ID << "," << data.dx  << "," << data.dy  << "," << data.dz << "," << data.status << endl;
        }
        //*/

        stream << "END" << endl;
    }
}
