#include "glviewcanvas.h"

#include <QColor>

// ViewListNode CLASS
ViewListNode::ViewListNode(int inZ, GLViewItem* inPtr)
{
	z=inZ;
	item=inPtr;
}

ViewListNode::~ViewListNode()
{
	if(item)
		delete item;
}

bool operator<(ViewListNode &left, ViewListNode &right)
{
	return left.z<right.z;
}

bool operator==(ViewListNode &left, ViewListNode &right)
{
	return left.z==right.z;
}


// canvas class
const GLfloat GLViewCanvas::mvRotStep=0.3f;

int GLViewCanvas::mvHeight = 0;
int GLViewCanvas::mvWidth = 0;

const GLfloat GLViewCanvas::mvNear = 0.1f;
const GLfloat GLViewCanvas::mvFar = 200;

GLViewCanvas::GLViewCanvas(QWidget* parent)
	:QGLWidget(parent)
{
	setFocusPolicy(Qt::StrongFocus);
	mvTimer = new QTimer(this);
	connect(mvTimer,SIGNAL(timeout()),this,SLOT(updateGL()));
	mvTimer->start(40); // about 25 FPS
	
	
	setAutoBufferSwap(false);
}

GLViewCanvas::~GLViewCanvas() {
	QListIterator<ViewListNode*> it(views);
	while(it.hasNext())
		delete it.next();
		
	views.clear();
}

void GLViewCanvas::initializeGL()
{
	glShadeModel(GL_SMOOTH);					// Enable Smooth Shading
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);				// Black Background
	glClearDepth(1.0f);						// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);					// Enables Depth Testing
	glDepthFunc(GL_LEQUAL);						// The Type Of Depth Testing To Do
	//   ShowCursor(false);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);


	glColor4f(1.0f, 1.0f, 1.0f, 0.5);				// Full Brightness.  50% Alpha
	
	mvCam.set(Point3(-1,0,0), Point3(0,0,0), Vector3(0,0,-1));
	
	mvCam.slide(-5.0,0.0,0.0); // choose a nicer initial position
}


void GLViewCanvas::resizeGL(int w, int h)
{
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h <= 0)
		h = 1;

	GLfloat ratio = 1.0f * w / h;
	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	mvHeight = h;
	mvWidth = w;

	// Set the viewport to be the entire window
	glViewport(0, 0, w, h);

	// Set the clipping volume
	mvCam.setShape(80,ratio,mvNear,mvFar);
	
	//mvCam.move(0.0,0.0,-5.0);
}

void GLViewCanvas::paintGL()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	//glLoadIdentity ( );

	

//	glTranslatef(0.0f,0.0f,z);
//	glRotatef(mvXRot,1.0f,0.0f,0.0f);
//	glRotatef(mvYRot,0.0f,1.0f,0.0f);

	glPushMatrix();
	
		// x
		QGLColor(Qt::red);
		glBegin(GL_LINES);
			glVertex3d(1,0,0);
			glVertex3d(0,0,0);
		glEnd();
		
		// y
		QGLColor(Qt::green);
		glBegin(GL_LINES);
			glVertex3d(0,1,0);
			glVertex3d(0,0,0);
		glEnd();
		
		// z
		QGLColor(Qt::blue);
		glBegin(GL_LINES);
			glVertex3d(0,0,1);
			glVertex3d(0,0,0);
		glEnd();
		
		QListIterator<ViewListNode*> it(views);
		while(it.hasNext()) {
			it.next()->item->drawGL();
		}

	glPopMatrix();

	swapBuffers();
}

void GLViewCanvas::mouseReleaseEvent(QMouseEvent *e)
{
	//qDebug("Mouse Up! (%d,%d,%d)",e->button(),e->x(),e->y());
	
	
	if(!mvDragFlag && mvStartDragPoint==e->pos())
	{
		// we have a click!
		//qDebug("Mouse Click!");
		
		mouseClick(e->button(), e->x(), e->y());
	}
	else
	{
		// we have a drag-stop!
		//qDebug("Mouse Drag Stop!");
		
		mouseStopDrag(e->button(), e->x(), e->y());
	}
}

void GLViewCanvas::mousePressEvent(QMouseEvent *e)
{
	//qDebug("Mouse Down! (%d,%d,%d)",e->button(),e->x(),e->y());
	
	mvDragState = e->button();
	mvStartDragPoint = e->pos(); //remember
	mvDragFlag=false;
}

void GLViewCanvas::mouseMoveEvent(QMouseEvent *e)
{
	//qDebug("Mouse Move! (%d,%d,%d)",e->button(),e->x(),e->y());

	if(!mvDragFlag && mvStartDragPoint!=e->pos())
	{
		mvDragFlag = true;
		
		// dispatch start drag
		//qDebug("Mouse Start Drag!");
		// go through the list with increasing z order
		mouseStartDrag(mvDragState, e->x(), e->y());
	}

		
	//qDebug("Mouse Drag!");
	// go through the list with increasing z order
	mouseDrag(e->button(), e->x(), e->y());
}

void GLViewCanvas::keyPressEvent(QKeyEvent *e)
{
	if(e->key() == Qt::Key_W ) // forward movement
		mvCam.slide(mvRotStep*.6,0.0,0.0);
	else if(e->key() == Qt::Key_A ) // left
		mvCam.slide(0.0,-mvRotStep*.6,0.0);
	else if(e->key() == Qt::Key_S ) // back
		mvCam.slide(-mvRotStep*.6,0.0,0.0);
	else if(e->key() == Qt::Key_D ) // right
		mvCam.slide(0.0,mvRotStep*.6,0.0);
	else
		e->ignore();
		
	//updateGL();
}

// default things handled by the canvas
void GLViewCanvas::mouseDrag(Qt::MouseButton, int x, int y)
{
	QPoint tempPoint;
	tempPoint.setX(x);
	tempPoint.setY(y);
	tempPoint-=mvStartRotate;

// 	qDebug("rotating... %d, %d",tempPoint.x(),tempPoint.y());

 	mvCam.pitch(mvRotStep * tempPoint.y()); //flip axes
 	mvCam.rotate(mvRotStep * tempPoint.x());
	
	mvStartRotate.setX(x);
	mvStartRotate.setY(y);
	
	//updateGL();
}

void GLViewCanvas::mouseClick(Qt::MouseButton, int, int)
{
	// default... nothing
}

void GLViewCanvas::mouseStartDrag(Qt::MouseButton, int x, int y)
{
	mvStartRotate.setX(x);
	mvStartRotate.setY(y);
}

void GLViewCanvas::mouseStopDrag(Qt::MouseButton, int, int)
{
	
}

void GLViewCanvas::addView(int z, GLViewItem* view) {
	ViewListNode* node = new ViewListNode(z,view);
	ViewListNode* other = NULL;
	QList<ViewListNode*>::iterator i;
	// do insertion sort
	
	if(views.count() == 0) {
		views.append(node);
	} else {
		for(i=views.begin(); i!=views.end(); i++) {
			other = *i;
			if( !((*other) < (*node)) ) {
				views.insert(i,node);
				return;
			}
		}
	}
}


// ********************* static utility functions ****************************

void GLViewCanvas::QGLColor(QColor inCol)
{
	glColor3f((float)inCol.red()/255.0,(float)inCol.green()/255.0,(float)inCol.blue()/255.0);
}
