#ifndef GRAPH_H_
#define GRAPH_H_

#include <QMap>
#include <QMapIterator>
#include <QUuid>

#include <QDataStream>
#include <QReadWriteLock>

class Graph;
class GraphNode;
class GraphLink;

class Graph
{
public:
	Graph();
	~Graph();

    uint nodeCount();
    uint linkCount();

    QMapIterator<QUuid, GraphNode*> nodeIterator();
    QMapIterator<QUuid, GraphLink*> linkIterator();

    GraphNode* insertNode(GraphNode* node);
    GraphLink* insertLink(GraphLink* link);

    GraphNode* getNode(const QUuid& uniqueID);
    GraphLink* getLink(const QUuid& uniqueID);
    GraphLink* getLink(GraphNode* fromNode, GraphNode* toNode);
    GraphLink* getLink(const QUuid& fromNodeID, const QUuid& toNodeID);

    bool hasNode(const QUuid& uniqueID);
    bool hasLink(const QUuid& uniqueID);
    bool hasLink(GraphNode* fromNode, GraphNode* toNode);
    bool hasLink(const QUuid& fromNodeID, const QUuid& toNodeID);

    void deleteNode(const QUuid& uniqueID);
    void deleteLink(const QUuid& uniqueID);

	friend QDataStream& operator<<(QDataStream&, Graph*);
	friend QDataStream& operator>>(QDataStream&, Graph*);
	
	// locking methods
	void acquireWriteLock() {
		lock.lockForWrite();
	}
	void acquireReadLock() {
		lock.lockForRead();
	}
	void releaseLock() {
		lock.unlock();
	}
	

private:
    QMap<QUuid, GraphNode*> nodes;
    QMap<QUuid, GraphLink*> links;
    
    QReadWriteLock lock;

};

QDataStream& operator<<(QDataStream&, Graph*);
QDataStream& operator>>(QDataStream&, Graph*);

class GraphReadLocker
{
private:
	Graph* graph;
public:
	GraphReadLocker(Graph* g) {
		graph = g;
		g->acquireReadLock();
	}
	
	~GraphReadLocker() {
		graph->releaseLock();
	}
};

class GraphWriteLocker
{
private:
	Graph* graph;
public:
	GraphWriteLocker(Graph* g) {
		graph = g;
		g->acquireWriteLock();
	}
	
	~GraphWriteLocker() {
		graph->releaseLock();
	}
};

class GraphNode
{
    friend class GraphLink;

public:
	GraphNode(Graph* graph, const QUuid& uniqueID);
	virtual ~GraphNode();

    QUuid getUniqueID();

    uint nodeCount();
    uint linkCount();

    QMapIterator<QUuid, GraphNode*> nodeIterator();
    QMapIterator<QUuid, GraphLink*> linkIterator();

    GraphLink* insertLink(const QUuid& uniqueID, GraphNode* toNode);
    GraphLink* insertLink(const QUuid& uniqueID, const QUuid& toNodeID);
    void deleteLink(GraphNode* toNode);
    void deleteLink(const QUuid& uniqueID);

    GraphLink* getLink(GraphNode* toNode);
    GraphLink* getLink(const QUuid& uniqueID);
    bool hasLink(GraphNode* toNode);
    bool hasLink(const QUuid& uniqueID);


protected:

    void addLink(GraphLink* link);
    void removeLink(GraphLink* link);

    Graph* graph;
    QUuid  uniqueID;

    QMap<QUuid, GraphNode*> nodes;
    QMap<QUuid, GraphLink*> links;
};

QDataStream& operator<<(QDataStream&, const GraphNode*);
QDataStream& operator>>(QDataStream&, GraphNode*);


class GraphLink
{
public:
	GraphLink(Graph* graph, const QUuid& uniqueID, GraphNode* fromNode, GraphNode* toNode);
	~GraphLink();

    QUuid      getUniqueID();
    GraphNode* getOtherNode(GraphNode* node);
    
    GraphNode* node1() { return fromNode; }
    GraphNode* node2() { return toNode; }

protected:
    Graph*     graph;
    QUuid      uniqueID;

    GraphNode* fromNode;
    GraphNode* toNode;
    QUuid      fromNodeID;
    QUuid      toNodeID;

};

QDataStream& operator<<(QDataStream&, const GraphLink*);
QDataStream& operator>>(QDataStream&, GraphLink*);


#endif 
