#ifndef MANIFOLD_H_
#define MANIFOLD_H_

#include <QUuid>
#include <QMap>
#include <QSet>
#include <QList>
#include <QSetIterator>

#include "Graph.h"
#include "ScanUtils.h"

#include "RFIDData.h"
#include "VictimRFIDData.h"
#include "ManifoldRenderer.h"



#ifdef USE_OPENGL
#include <QtOpenGL>
#include <QColor>
#endif

#ifdef USE_QT_DRAWING
#include <QPainter>
#endif

class Manifold;
class ManifoldNode;
class ManifoldLink;



class Manifold : public Graph
{

private:
    static const float SUBMAP_MAXTRANSLATIONVARIANCE;
    static const float SUBMAP_MAXROTATIONVARIANCE;

public:
	Manifold();
	virtual ~Manifold();

    void insertPatch(ScanObservation* patch);
    void insertRelation(ScanObservation* fromPatch, ScanObservation* toPatch, Matrix3 mcov);
    
    void insertRelation(const QUuid& fromPatch, const QUuid& toPatch, Matrix3 mcov);

    Manifold* getSubMap(ScanObservation* patch);
    Manifold* getConnectedSubSet();
    Manifold* getConnectedSubSet(ManifoldNode* node);

	Bounds computeBounds();
	
	void insertStartNode(QUuid);

    bool hasRFID(const QString& ID);
    bool hasVictim(const QString& ID);

	void updateVictimStatus(QString ID, QString status);
	QString getVictimStatus(QString ID) {
		return victimStatus.value(ID);
	}

	QList< RFIDData >       getRFIDs();
    QList< QUuid >          getRFIDPatches(QString ID);
	QList< VictimRFIDData > getVictimRFIDs();
    QList< QUuid >          getVictimPatches(QString ID);

	QSetIterator<QUuid>     startNodeIterator() {
		return QSetIterator<QUuid>(startNodes);
	}

	void render(ManifoldRenderer* renderer);

private:

    void getSubMap(Manifold* submap, QList<QUuid>& done, Matrix3 sumcov, ManifoldNode* node);
    void getConnectedSubSet(Manifold* subset, QList<QUuid>& done, ManifoldNode* node);

    QMap< QString, QList<QUuid> > rfidPatches;
	QMap< QString, QList<QUuid> > victimPatches;
	
	QMap< QString, QString> victimStatus;
	
	QSet< QUuid > startNodes;


// for current 2D view
#ifdef USE_QT_DRAWING
	void drawQt(QPainter&);
#endif

// for potential 3D view
#ifdef USE_OPENGL
	void drawGL();
	void setColors(QColor, QColor);
#endif

};

QDataStream& operator>>(QDataStream&, Manifold&);
QDataStream& operator<<(QDataStream&, const Manifold&);


class ManifoldNode : public GraphNode
{
    friend class Manifold;

public:
	ManifoldNode(Manifold* manifold, ScanObservation* patch);
	~ManifoldNode();

    ScanObservation* patch;
	void moveBy(Vector2 center, Pose2D dpose);
	void moveAllConnectedBy(Vector2 center, Pose2D dpose);

private:
    void copyTo(Manifold* manifold);
};




class ManifoldLink : public GraphLink
{
    friend class Manifold;

public:
	ManifoldLink(Manifold* manifold, ManifoldNode* fromNode, ManifoldNode* toNode, const Matrix3& mcov);
	~ManifoldLink();

    Matrix3 mcov;

private:
    void copyTo(Manifold* manifold);

};

#endif
