#ifndef UTIL3D_H_
#define UTIL3D_H_

#include <QtDebug>
#include <QDataStream>

#include "Exception.h"

#include <math.h>

// max: This gives an error for me, they are already defined in math.h
#ifdef ON_WIN32
const double M_PI = 3.14159265358979323846;
const double M_PI_2 = M_PI / 2;
#endif

struct Pose3D {
	float x;
	float y;
	float z;
	float rotX;
	float rotY;
	float rotZ;
};

QDataStream& operator<<(QDataStream&, const Pose3D&);
QDataStream& operator>>(QDataStream&, Pose3D&);

Pose3D operator-(const Pose3D&, const Pose3D&);
Pose3D operator+(const Pose3D&, const Pose3D&);

class Vector3;

class Matrix4 {
public:
	float nums[16];
	Matrix4(float m[16]) {
		for(int i=0; i<16; i++)
			nums[i] = m[i];
	}
	Matrix4() {
		//qDebug() << "Matrix4 no-arg";
		loadIdentity();	
	}
	Matrix4(const Matrix4& m) {
		for(int i=0; i<16; i++)
			nums[i] = m.nums[i];
	}
	

	void loadIdentity();
	
	// use these for consecutive actions
	void translate(float, float, float);
	void rotate(float, float, float);
	void invRotate(float, float, float);
	void invTranslate(float, float, float);
	
	// one shot setters
	void setRotation(float rx, float ry, float rz);
	void setInvRotation(float rx, float ry, float rz);
	void setTranslation(float x, float y, float z);
	
	void fromPose3D(const Pose3D& pose) {
		translate(pose.x,pose.y,pose.z);
		rotate(pose.rotX,pose.rotY,pose.rotZ);
	}
	
	void invFromPose3D(const Pose3D& pose) {
		invRotate(pose.rotX,pose.rotY,pose.rotZ);
		invTranslate(pose.x,pose.y,pose.z);
	}
	
	float& operator()(int i, int j) {
		//ith row, jth col
		if(i>=4 || j>=4)
			throw Exception("Out of bounds!");

		return nums[ index(i,j,4) ];
	}
	const float& operator()(int i, int j) const {
		//ith row, jth col
		if(i>=4 || j>=4)
			throw Exception("Out of bounds!");

		return nums[ index(i,j,4) ];
	}
	
	Matrix4 transpose();
	Matrix4 inverse();
	
	Vector3 multiply(const Vector3 &);
	Matrix4 multiply(const Matrix4 &);
	
	Matrix4 &operator=(const Matrix4&);
	
	Matrix4  operator-(const Matrix4&);
	Matrix4  operator+(const Matrix4&);
	Matrix4  operator*(const Matrix4&);
	Vector3  operator*(const Vector3&);
	
private:

	int index(int i, int j, int size) const {
		return i*size+j;
	}
};

Vector3 operator*(const Vector3 & v, const Matrix4 & m);
Matrix4 operator*(const float & f, const Matrix4 & m);

// *****************************************************************************************
// * FROM the book "Computer Graphics using Open GL", by F.S. Hill, Jr., Pearson Education *
// *****************************************************************************************

//@@@@@@@@@@@@@@@@@@ Point3 class @@@@@@@@@@@@@@@@
class Point3{ 
public: 
	float x,y,z;
	void set(float dx, float dy, float dz){x = dx; y = dy; z = dz;}
	void set(Point3& p){x = p.x; y = p.y; z = p.z;} 	
	Point3(float xx, float yy, float zz){x = xx; y = yy; z = zz;}
	Point3(){x = y = z = 0;}
	void build4tuple(float v[])
	{// load 4-tuple with this color: v[3] = 1 for homogeneous
		v[0] = x; v[1] = y; v[2] = z; v[3] = 1.0f;
	}
};
//@@@@@@@@@@@@@@@@@@ Vector3 class @@@@@@@@@@@@@@@@
class Vector3{ 
public: 
	float x,y,z;
	void set(float dx, float dy, float dz){ x = dx; y = dy; z = dz;} 
	void set(const Vector3& v){ x = v.x; y = v.y; z = v.z;}
	void flip(){x = -x; y = -y; z = -z;} // reverse this vector
	void setDiff(Point3& a, Point3& b)//set to difference a - b
	{ x = a.x - b.x; y = a.y - b.y; z = a.z - b.z;}
	void normalize()//adjust this vector to unit length
	{
		double sizeSq = x * x + y * y + z * z;
		if(sizeSq < 0.0000001)
		{
			//cerr << "\nnormalize() sees vector (0,0,0)!";
			return; // does nothing to zero vectors;
		}
		float scaleFactor = 1.0/(float)sqrt(sizeSq);
		x *= scaleFactor; y *= scaleFactor; z *= scaleFactor;
	}
	Vector3(float xx, float yy, float zz){x = xx; y = yy; z = zz;}
	Vector3(float xx, float yy){x = xx; y = yy; z = 0.0f;}
	Vector3(const Vector3& v){x = v.x; y = v.y; z = v.z;}
	Vector3(){x = y = z = 0;} //default constructor
	Vector3 cross(Vector3 b) //return this cross b
	{
		Vector3 c(y*b.z - z*b.y, z*b.x - x*b.z, x*b.y - y*b.x);
		return c;
	}
	float dot(Vector3 b) const // return this dotted with b
	{return x * b.x + y * b.y + z * b.z;}
	
	float fmin();
	float fmax();
	
	char minIndex();
	
	float length() const;
	
	Vector3 operator-(const Vector3 & right);
	Vector3 operator+(const Vector3 & right);
	Vector3& operator=(const Vector3 & right);
	
	Vector3 operator/(const Vector3 & right);
	Vector3 operator/(const float & right);
	
private:
	float divide(float nom, float denom);
};
Vector3 operator*(const float & left, const Vector3 & right);
float operator*(const Vector3& left, const Vector3 & right);

#ifdef USE_OPENGL
//@@@@@@@@@@@@@@@@@@@@@@@@@ Camera class @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Camera{
  private:
	Point3 eye;
	Vector3 u,v,n;
	double viewAngle, aspect, nearDist, farDist; // view volume shape
	void rotateZ(Vector3& vec, float angle);

  public:
	Camera(); // default constructor
	void set(Point3 eye, Point3 look, Vector3 up); // like gluLookAt()
	void roll(float angle); // roll it
	void pitch(float angle); // increase pitch
	void yaw(float angle); // yaw it
	void slide(float , float , float ); // slide it
	void setShape(float vAng, float asp, float nearD, float farD);
	
	void setModelViewMatrix(); // tell OpenGL where the camera is
	
	// my additions
	void setEye(float , float , float );
	void move(float, float, float);
	void rotate(float); //< will yaw the camera without a roll!
};
#endif /* USE_OPENGL */


#endif /*UTIL3D_H_*/
