#include "XSocket.h"

#include <errno.h>

#include <QtDebug>
#include <QDataStream>

const int XSocket::BUFSIZE = 100 * 1024; // 100kB

XSocket::XSocket(SOCKET existingSocket)
{
	int error = WSAStartup (0x0202,&wsa);
	if (error)
	{
		throw Exception("Error:  You need WinSock 2.2!", __FILE__, __LINE__);
	}

	sockfd = existingSocket;
	aIsConnected = true;
    
    buf = new char[BUFSIZE]; // 100kB
    alreadyReadBefore = 0;
    
    queueWhenDisconnected = false;
}

XSocket::XSocket()
{
	int error = WSAStartup (0x0202,&wsa);
	if (error)
	{
		throw Exception("Error:  You need WinSock 2.2!", __FILE__, __LINE__);
	}

	if ((sockfd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		WSACleanup ();
        throw Exception("Could not make a new socket! " + lastError(), __FILE__, __LINE__);
    }
    aIsConnected = false;
    
    buf = new char[BUFSIZE]; // 100kB
    alreadyReadBefore = 0;
    
    queueWhenDisconnected = false;
}

XSocket::~XSocket()
{
	disconnect();
	WSACleanup();
	delete buf;
}

void XSocket::setQueueWhenDisconnected(bool queue)
{
	queueWhenDisconnected = queue;
}

void XSocket::connect(const QString& host, uint port)
{
	remoteAddr.sin_family = AF_INET;    // host byte order 
    remoteAddr.sin_port = htons(port);  // short, network byte order 
    remoteAddr.sin_addr.s_addr = inet_addr(host.toAscii());
    
    if( ::connect(sockfd, (sockaddr*)&remoteAddr, sizeof(remoteAddr)) == -1 )
  	{
		aIsConnected = false;
  		throw Exception("Couldn't connect to "+host+"! " + lastError(), __FILE__, __LINE__ );
  	}
  	
	aIsConnected = true;
	
	int sent=0;
	while( !sendWhenDisconnectedQueue.isEmpty() )
	{
		sent = sendAll( sendWhenDisconnectedQueue.dequeue() );
		qDebug() << "sent queued message, size: " << sent;
	}
}

void XSocket::disconnect()
{
	if( aIsConnected ) {
		closesocket(sockfd);
		aIsConnected = false;
	}
}

int XSocket::sendAll(const QByteArray& in)
{
	if( !aIsConnected && !queueWhenDisconnected )
		throw Exception("Socket not connected.", __FILE__, __LINE__);
		
	if( !aIsConnected && queueWhenDisconnected )
	{
		sendWhenDisconnectedQueue.enqueue(in);
		return 0;
	}
	
	int numWrittenEver = 0;
	int numWritten = 0;
	const char* charData = in.data();;
	int charDataLen = in.size();
	
	while( numWrittenEver < charDataLen )
	{
		if( ( numWritten = ::send( sockfd, charData+numWrittenEver, charDataLen-numWrittenEver, 0) ) == -1 )
		{
			aIsConnected = false;
			throw Exception("Couldn't send data! " + lastError() , __FILE__, __LINE__);
		}
		numWrittenEver += numWritten;
	}
	
	return numWrittenEver;
}

int XSocket::send(const QString& data)
{
	QString toBeSent = data;
	if( !toBeSent.endsWith("\r\n") )
	{
		toBeSent.append("\r\n");
	}
	
	return sendAll(toBeSent.toAscii());
}

int XSocket::sendBytes(const QByteArray& data)
{
	QByteArray env;
	
	{
		QDataStream out(&env, QIODevice::WriteOnly);
		out << data;
	}
	
	return sendAll(env);
}

QByteArray XSocket::receiveBytes(unsigned long msec)
{
	if( !aIsConnected )
		throw Exception("Socket not connected.", __FILE__, __LINE__);
		
	if( !receiveBytesQueue.isEmpty() )
		return receiveBytesQueue.dequeue();
	
	if( !waitForData(msec) )
		return QByteArray(); // nothing to read
	
	readMore();
	
	bool firstPass = true;
	while(true)
	{
		int toBeRead = 0;
		if( alreadyReadBefore >= sizeof(quint32) )
		{
			QByteArray size = QByteArray::fromRawData(buf, sizeof(quint32));
			QDataStream in(&size, QIODevice::ReadOnly);
			quint32 len;
			in >> len;
			toBeRead = len;
		}
		else
		{
			break;
		}
		
		while( firstPass && toBeRead+sizeof(quint32) > alreadyReadBefore )
			readMore();
			
		QByteArray newArray( buf+sizeof(quint32), toBeRead ); // deep copy data
		
		receiveBytesQueue.enqueue(newArray);
		
		alreadyReadBefore = copyToFrontOfBuffer(toBeRead+sizeof(quint32),alreadyReadBefore);
		
		firstPass = false;
	}
	
	if(receiveBytesQueue.isEmpty())
		return QByteArray();
	else
		return receiveBytesQueue.dequeue();
	
}

int XSocket::findNextCRLF(char* buf, int len, int offset)
{
	char previous = 0;
	int start = offset;
	for(int i=start; i<len; i++) {
		if( previous == '\r' && buf[i] == '\n' ) // found a newline thing
		{
			return i+1;
		}
		previous = buf[i];
	}

	return -1;
}

int XSocket::copyToFrontOfBuffer(int offset, int end)
{
	for(int i=offset; i<end; i++) {
		buf[i-offset] = buf[i];
	}
	buf[end-offset] = 0; //null-terminate
	return end-offset; // next byte available for writing
}

bool XSocket::isConnected() {
	return aIsConnected;
}

bool XSocket::waitForData(unsigned long msec)
{
	fd_set sock_set;
	FD_ZERO(&sock_set);
	FD_SET( sockfd, &sock_set);
	timeval timeout;
	timeout.tv_sec = msec/1000;
	timeout.tv_usec = msec%1000 * 1000;
	
	int sel = select( sockfd+1,  &sock_set, NULL, NULL, &timeout );
	if( sel == -1 )
	{
		aIsConnected = false;
		throw Exception("Couldn't read data, select failed! " + lastError(), __FILE__, __LINE__ );
	}

	return !( sel == 0 || !FD_ISSET(sockfd, &sock_set) );
}

void XSocket::readMore()
{
	int numNewlyRead = 0;
	if( ( numNewlyRead = recv(sockfd, buf+alreadyReadBefore, BUFSIZE-1-alreadyReadBefore, 0 ) ) == -1 )
	{
		aIsConnected = false;
		throw Exception("Couldn't read data! " + lastError() , __FILE__, __LINE__);
	}
	alreadyReadBefore += numNewlyRead;
	
	buf[alreadyReadBefore] = 0;
}

QString XSocket::receive(unsigned long msec)
{
	if( !aIsConnected )
		throw Exception("Socket not connected.", __FILE__, __LINE__);
		
	if( !receiveQueue.isEmpty() )
		return receiveQueue.dequeue();
		
		
	if( !waitForData(msec) )
		return QString(); // nothing to read
	
	readMore();
	
	// we might have gotten so much data that we read multiple lines, check!
	
	int start = 0;
	int next = 0;
	
	if( alreadyReadBefore == 0 ) // nothing read so far, assume there isn't any more for now
		return QString();
	
	next = findNextCRLF(buf, alreadyReadBefore, start);
	while( next == -1 ) // we have to read again!
	{
		readMore();
		
		next = findNextCRLF(buf, alreadyReadBefore, start);
	}
	
	// we have at least one whole line in the buffer now
	
	while( next != -1 )
	{
		buf[next-2] = 0;
		receiveQueue.enqueue(QString(buf+start));
		start = next;
		
		if( next == alreadyReadBefore ) // we are done
		{
			alreadyReadBefore = 0;
			break;
		}
		
		next = findNextCRLF(buf, alreadyReadBefore, next);
	}
	
	if( next == -1 ) // we have an unfinished line left in the buffer
	{
		alreadyReadBefore = copyToFrontOfBuffer(start,alreadyReadBefore);
	}
	
	// there is always at least one string in the queue now since all data sent
	// by the server ends in \r\n and we don't cut it off before
	
	return receiveQueue.dequeue();
}

QString XSocket::lastError()
{
	switch( WSAGetLastError() )
	{
		case WSAEINTR: return "Interrupted function call.";
		case WSAEACCES: return "Permission denied.";
		case WSAEFAULT: return "Bad address.";
		case WSAEINVAL: return "Invalid argument.";
		case WSAEMFILE: return "Too many open files.";
		case WSAEWOULDBLOCK: return "Resource temporarily unavailable.";
		case WSAEINPROGRESS: return "Operation now in progress.";
		case WSAEALREADY: return "Operation already in progress.";
		case WSAENOTSOCK: return "Socket operation on nonsocket.";
		case WSAEDESTADDRREQ: return "Destination address required.";
		case WSAEMSGSIZE: return "Message too long.";
		case WSAEPROTOTYPE: return "Protocol wrong type for socket.";
		case WSAENOPROTOOPT: return "Bad protocol option.";
		case WSAEPROTONOSUPPORT: return "Protocol not supported.";
		case WSAESOCKTNOSUPPORT: return "Socket type not supported.";
		case WSAEOPNOTSUPP: return "Operation not supported.";
		case WSAEPFNOSUPPORT: return "Protocol family not supported.";
		case WSAEAFNOSUPPORT: return "Address family not supported by protocol family.";
		case WSAEADDRINUSE: return "Address already in use.";
		case WSAEADDRNOTAVAIL: return "Cannot assign requested address.";
		case WSAENETDOWN: return "Network is down.";
		case WSAENETUNREACH: return "Network is unreachable.";
		case WSAENETRESET: return "Network dropped connection on reset.";
		case WSAECONNABORTED: return "Software caused connection abort.";
		case WSAECONNRESET: return "Connection reset by peer.";
		case WSAENOBUFS: return "No buffer space available.";
		case WSAEISCONN: return "Socket is already connected.";
		case WSAESHUTDOWN: return "Cannot send after socket shutdown.";
		case WSAETIMEDOUT: return "Connection timed out.";
		case WSAECONNREFUSED: return "Connection refused.";
		case WSAEHOSTDOWN: return "Host is down.";
		case WSAEHOSTUNREACH: return "No route to host.";
		case WSAEPROCLIM: return "Too many processes.";
		case WSASYSNOTREADY: return "Network subsystem is unavailable.";
		case WSAVERNOTSUPPORTED: return "Winsock.dll version out of range.";
		case WSANOTINITIALISED: return "Successful WSAStartup not yet performed.";
		case WSAEDISCON: return "Graceful shutdown in progress.";
		case WSATYPE_NOT_FOUND: return "Class type not found.";
		case WSAHOST_NOT_FOUND: return "Host not found.";
		case WSATRY_AGAIN: return "Nonauthoritative host not found.";
		case WSANO_RECOVERY: return "This is a nonrecoverable error.";
		case WSANO_DATA: return "Valid name, no data record of requested type.";
		case WSA_INVALID_HANDLE: return "Specified event object handle is invalid.";
		case WSA_INVALID_PARAMETER: return "One or more parameters are invalid.";
		case WSA_IO_INCOMPLETE: return "Overlapped I/O event object not in signaled state.";
		case WSA_IO_PENDING: return "Overlapped operations will complete later.";
		case WSA_OPERATION_ABORTED: return "Overlapped operation aborted.";
		case WSASYSCALLFAILURE: return "System call failure.";
		default: return QString("Unknown error: ").append(WSAGetLastError());
	};
}
