#include "UsarServer.h"

#include <QImage>

#include "QtSimpleManifoldRenderer.h"

UsarServer::UsarServer(uint port)
	:Server(port)
{
	manifold = new Manifold();
}

UsarServer::~UsarServer()
{
	delete manifold;
}

ServerConnection* UsarServer::newServerConnection(XSocket* sock)
{
	return new UsarConnection(sock, this);
}

void UsarServer::mergeIslands() {
	GraphWriteLocker lock(manifold);

	WeightedScanMatcher matcher;
	ManifoldSlam slam(&matcher,manifold);

	QSetIterator<QUuid> startNodes = this->manifold->startNodeIterator();
	QList< QUuid > nodes;
	while(startNodes.hasNext()) {
		nodes.append(startNodes.next());
	}

	if (nodes.count() < 2) {
		return;
	}

	QList< QUuid > done;
	QUuid target;
	QUuid source;
	
	done.append(nodes[0]);
	
	for (int i = 0; i < nodes.count(); i++) {
		target = nodes[i];
		if (done.contains(target)) {
			for (int j = 0; j < nodes.count(); j++) {
				source = nodes[j];
				if (done.contains(source)) {
					continue;
				}
				if (slam.mergeIslands(source, target)) {
					qDebug() << QString("-- Merged Islands %1 and %2").arg(i).arg(j);
					done.append(source);
				}
			}
		}
	}
}

void UsarServer::saveMap(int pxPerMeter, QString filename)
{
	if (!filename.toLower().endsWith(".png"))
        filename.append(".png");

	//Manifold* subset = this->manifold->getConnectedSubSet();
	Manifold* subset = this->manifold;
	//Bounds bounds    = subset->computeBounds();
	Bounds bounds    = this->manifold->computeBounds();

    qDebug() << "SAVING MAP ...";
    qDebug() << QString("-- Map bounds are (%1,%2 %3,%4)").arg(bounds.top).arg(bounds.left).arg(bounds.height).arg(bounds.width);
    qDebug() << QString("-- Full Manifold has %1 nodes and %2 links").arg(this->manifold->nodeCount()).arg(this->manifold->linkCount());
    qDebug() << QString("-- Connected subset has %1 nodes and %2 links").arg(subset->nodeCount()).arg(subset->linkCount());

	int iHeight = (float)pxPerMeter * bounds.height / 1000;
	int iWidth  = (float)pxPerMeter * bounds.width / 1000;

	qDebug() << QString("-- Making image of size (%1,%2)").arg(iWidth).arg(iHeight);

	// keep aspect ratio of bounds
    QImage img = QImage(iWidth, iHeight, QImage::Format_ARGB32_Premultiplied);
    
    // make sure maprenderer gets destroyed so the painter gets destroyed too
    {
		ManifoldRenderer* renderer = NULL;

		// switch it here?
		//renderer = new QtSimpleManifoldRenderer(&img, bounds);
		renderer = new QtManifoldRenderer(&img, bounds);
		
		if (renderer != NULL) {
			//this->manifold->render(renderer);
			subset->render(renderer);
			delete renderer;
			renderer = NULL;
		}
    }

    if (img.save(filename, "PNG")) {
        qDebug() << QString("-- Map saved to %1").arg(filename);
        qDebug() << QString("-- Done");
    } else {
        qDebug() << QString("-- Could not save map to %1").arg(filename);
    }

	//delete subset;
}

void UsarServer::saveRFIDs(QString filename)
{
	QFile file(filename);
	if (file.open(QIODevice::WriteOnly | QFile::Truncate)) {

        qDebug() << "Saving RFIDs to file: " << filename;

		// shoud be doing breadth first
		
        QTextStream stream(&file);
        
        
        QList<RFIDData> rfids;
        {
        		GraphReadLocker lock(manifold);
       		rfids = this->manifold->getRFIDs();
        }
        
        for(int i=0; i<rfids.size(); i++)
        {
    		qDebug() << "- " << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0;
    		stream << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0 << endl;
        }
        
        /*
        QMapIterator< QString, RFIDObservation* > rfidIterator = this->manifold->rfidIterator();
        RFIDObservation* rfid;
        RFIDData data;
        while (rfidIterator.hasNext()) {
            rfidIterator.next();
            rfid = rfidIterator.value();
            data = rfid->getRFID();
            stream << data.ID << "," << data.dx  << "," << data.dy  << "," << data.dz << endl;
        }
        //*/

        stream << "END" << endl;
	}	
}

void UsarServer::saveVictims(QString filename)
{
	QFile file(filename);
	if (file.open(QIODevice::WriteOnly | QFile::Truncate)) {
		
        qDebug() << "Saving Victims to file: " << filename;

		
	    QTextStream stream(&file);

		QList<VictimRFIDData> rfids;
		{
			GraphReadLocker lock(manifold);
			rfids = this->manifold->getVictimRFIDs();
		}
        for(int i=0; i<rfids.size(); i++)
        {
        		qDebug() << "- " << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0 << "," << this->manifold->getVictimStatus( rfids[i].ID );
        		stream << rfids[i].ID << "," << rfids[i].dx  << "," << rfids[i].dy  << "," << 0 << "," << this->manifold->getVictimStatus( rfids[i].ID ) << endl;
        }
		//*/

		/*
        QMapIterator< QString, VictimObservation* > victimIterator = this->manifold->victimIterator();
        VictimObservation* victim;
        VictimRFIDData data;
        while (victimIterator.hasNext()) {
            victimIterator.next();
            victim = victimIterator.value();
            data   = victim->getRFID();
            stream << data.ID << "," << data.dx  << "," << data.dy  << "," << data.dz << "," << data.status << endl;
        }
        //*/

        stream << "END" << endl;
    }
}

