#include <QtGui>

#include "MainWindow.h"
#include "MapWindow.h"

MainWindow::MainWindow()
{
    workspace = new QWorkspace;
    setCentralWidget(workspace);
    connect(workspace, SIGNAL(windowActivated(QWidget *)),
            this, SLOT(updateMenus()));
    windowMapper = new QSignalMapper(this);
    connect(windowMapper, SIGNAL(mapped(QWidget *)),
            workspace, SLOT(setActiveWindow(QWidget *)));

    createActions();
    createMenus();
    createToolBars();
    createStatusBar();
    updateMenus();

    readSettings();

    setWindowTitle(tr("UvA Rescue Team - Scanmatcher"));
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    workspace->closeAllWindows();
    if (activeMapWindow()) {
        event->ignore();
    } else {
        writeSettings();
        event->accept();
    }
}

void MainWindow::open()
{
    QString fileName = QFileDialog::getOpenFileName(this);
    if (!fileName.isEmpty()) {
        MapWindow *existing = findMapWindow(fileName);
        if (existing) {
            workspace->setActiveWindow(existing);
            return;
        }

        MapWindow *child = createMapWindow();
        child->show();

        if (fileName.endsWith(".log") || fileName.endsWith(".dat")) {
            if (child->loadLogFile(fileName)) {
                statusBar()->showMessage(tr("Log file loaded"), 2000);
            }
        } else if (child->loadImageFile(fileName)) {
            statusBar()->showMessage(tr("Image file loaded"), 2000);
        }
    }
}

void MainWindow::save()
{
    if (activeMapWindow()->save())
        statusBar()->showMessage(tr("File saved"), 2000);
}

void MainWindow::saveAs()
{
    if (activeMapWindow()->saveAs())
        statusBar()->showMessage(tr("File saved"), 2000);
}

void MainWindow::useNoSM() {
    activeMapWindow()->setScanMatcher(new ScanMatcher());
}

void MainWindow::useWeightedSM() {
    activeMapWindow()->setScanMatcher(new WeightedScanMatcher());
}

void MainWindow::useUnweightedSM() {
    activeMapWindow()->setScanMatcher(new UnweightedScanMatcher());
}

void MainWindow::zoomIn() {
    activeMapWindow()->zoomBy(1.25);
}

void MainWindow::zoomOut() {
    activeMapWindow()->zoomBy(0.75);
}

void MainWindow::zoomReset()
{
    activeMapWindow()->zoomTo(1.0);
}


void MainWindow::about()
{
   QMessageBox::about(this, tr("UvA Rescue Team - Scanmatcher"),
            tr("The <b>Scanmatcher</b> application allows you to test the "
               "scanmatcher software on pre-recorded logfiles."));
}

void MainWindow::updateMenus()
{
    bool hasMapWindow = (activeMapWindow() != 0);

    separatorAct->setVisible(hasMapWindow);

    saveAct->setEnabled(hasMapWindow);
    saveAsAct->setEnabled(hasMapWindow);
    closeAct->setEnabled(hasMapWindow);
    closeAllAct->setEnabled(hasMapWindow);

    tileAct->setEnabled(hasMapWindow);
    cascadeAct->setEnabled(hasMapWindow);
    arrangeAct->setEnabled(hasMapWindow);
    nextAct->setEnabled(hasMapWindow);
    previousAct->setEnabled(hasMapWindow);

    zoomInAct->setEnabled(hasMapWindow);
    zoomOutAct->setEnabled(hasMapWindow);
    zoomResetAct->setEnabled(hasMapWindow);
}

void MainWindow::updateWindowMenu()
{
    windowMenu->clear();
    windowMenu->addAction(closeAct);
    windowMenu->addAction(closeAllAct);
    windowMenu->addSeparator();
    windowMenu->addAction(tileAct);
    windowMenu->addAction(cascadeAct);
    windowMenu->addAction(arrangeAct);
    windowMenu->addSeparator();
    windowMenu->addAction(nextAct);
    windowMenu->addAction(previousAct);
    windowMenu->addAction(separatorAct);

    QList<QWidget *> windows = workspace->windowList();
    separatorAct->setVisible(!windows.isEmpty());

    for (int i = 0; i < windows.size(); ++i) {
        MapWindow *child = qobject_cast<MapWindow *>(windows.at(i));

        QString text;
        if (i < 9) {
            text = tr("&%1 %2").arg(i + 1)
                               .arg(child->userFriendlyCurrentFile());
        } else {
            text = tr("%1 %2").arg(i + 1)
                              .arg(child->userFriendlyCurrentFile());
        }
        QAction *action  = windowMenu->addAction(text);
        action->setCheckable(true);
        action ->setChecked(child == activeMapWindow());
        connect(action, SIGNAL(triggered()), windowMapper, SLOT(map()));
        windowMapper->setMapping(action, child);
    }
}

MapWindow *MainWindow::createMapWindow()
{
    MapWindow *child = new MapWindow;
    workspace->addWindow(child);
    return child;
}

void MainWindow::createActions()
{
    separatorAct = new QAction(this);
    separatorAct->setSeparator(true);


    openAct = new QAction(QIcon(":/images/open.png"), tr("&Open..."), this);
    openAct->setShortcut(tr("Ctrl+O"));
    openAct->setStatusTip(tr("Open an existing file"));
    connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

    saveAct = new QAction(QIcon(":/images/save.png"), tr("&Save"), this);
    saveAct->setShortcut(tr("Ctrl+S"));
    saveAct->setStatusTip(tr("Save the document to disk"));
    connect(saveAct, SIGNAL(triggered()), this, SLOT(save()));

    saveAsAct = new QAction(tr("Save &As..."), this);
    saveAsAct->setStatusTip(tr("Save the document under a new name"));
    connect(saveAsAct, SIGNAL(triggered()), this, SLOT(saveAs()));

    exitAct = new QAction(tr("E&xit"), this);
    exitAct->setShortcut(tr("Ctrl+Q"));
    exitAct->setStatusTip(tr("Exit the application"));
    connect(exitAct, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));

    closeAct = new QAction(tr("Cl&ose"), this);
    closeAct->setShortcut(tr("Ctrl+F4"));
    closeAct->setStatusTip(tr("Close the active window"));
    connect(closeAct, SIGNAL(triggered()), workspace, SLOT(closeActiveWindow()));

    closeAllAct = new QAction(tr("Close &All"), this);
    closeAllAct->setStatusTip(tr("Close all the windows"));
    connect(closeAllAct, SIGNAL(triggered()), workspace, SLOT(closeAllWindows()));


    useNoSMAct = new QAction(tr("Disable ScanMatcher"), this);
    useNoSMAct->setStatusTip(tr("Disable ScanMatcher"));
    useNoSMAct->setCheckable(true);
    useNoSMAct->setChecked(false);
    connect(useNoSMAct, SIGNAL(triggered()), this, SLOT(useNoSM()));

    useWeightedSMAct = new QAction(tr("Use &Weighted ScanMatcher"), this);
    useWeightedSMAct->setStatusTip(tr("Switch to Weighted ScanMatcher"));
    useWeightedSMAct->setCheckable(true);
    useWeightedSMAct->setChecked(true);
    connect(useWeightedSMAct, SIGNAL(triggered()), this, SLOT(useWeightedSM()));

    useUnweightedSMAct = new QAction(tr("Use &Unweighted ScanMatcher"), this);
    useUnweightedSMAct->setStatusTip(tr("Switch to Unweighted ScanMatcher"));
    useUnweightedSMAct->setCheckable(true);
    useUnweightedSMAct->setChecked(false);
    connect(useUnweightedSMAct, SIGNAL(triggered()), this, SLOT(useUnweightedSM()));



    zoomInAct = new QAction(QIcon(":/images/zoomin.png"), tr("Zoom &In (25%)"), this);
    zoomInAct->setShortcut(tr("Ctrl++"));
    zoomInAct->setStatusTip(tr("Zoom in"));
    connect(zoomInAct, SIGNAL(triggered()), this, SLOT(zoomIn()));

    zoomOutAct = new QAction(QIcon(":/images/zoomout.png"), tr("Zoom &Out (25%)"), this);
    zoomOutAct->setShortcut(tr("Ctrl+-"));
    zoomOutAct->setStatusTip(tr("Zoom out"));
    connect(zoomOutAct, SIGNAL(triggered()), this, SLOT(zoomOut()));

    zoomResetAct = new QAction(QIcon(":/images/zoom.png"), tr("&Normal Size"), this);
    zoomResetAct->setShortcut(tr("Ctrl+R"));
    zoomResetAct->setStatusTip(tr("Reset zoom"));
    connect(zoomResetAct, SIGNAL(triggered()), this, SLOT(zoomReset()));



    tileAct = new QAction(tr("&Tile"), this);
    tileAct->setStatusTip(tr("Tile the windows"));
    connect(tileAct, SIGNAL(triggered()), workspace, SLOT(tile()));

    cascadeAct = new QAction(tr("&Cascade"), this);
    cascadeAct->setStatusTip(tr("Cascade the windows"));
    connect(cascadeAct, SIGNAL(triggered()), workspace, SLOT(cascade()));

    arrangeAct = new QAction(tr("Arrange &icons"), this);
    arrangeAct->setStatusTip(tr("Arrange the icons"));
    connect(arrangeAct, SIGNAL(triggered()), workspace, SLOT(arrangeIcons()));

    nextAct = new QAction(tr("Ne&xt"), this);
    nextAct->setShortcut(tr("Ctrl+F6"));
    nextAct->setStatusTip(tr("Move the focus to the next window"));
    connect(nextAct, SIGNAL(triggered()), workspace, SLOT(activateNextWindow()));

    previousAct = new QAction(tr("Pre&vious"), this);
    previousAct->setShortcut(tr("Ctrl+Shift+F6"));
    previousAct->setStatusTip(tr("Move the focus to the previous window"));
    connect(previousAct, SIGNAL(triggered()), workspace, SLOT(activatePreviousWindow()));



    aboutAct = new QAction(tr("&About"), this);
    aboutAct->setStatusTip(tr("Show the application's About box"));
    connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));
}

void MainWindow::createMenus()
{
    fileMenu = menuBar()->addMenu(tr("&File"));
    fileMenu->addAction(openAct);
    fileMenu->addAction(saveAct);
    fileMenu->addAction(saveAsAct);
    fileMenu->addSeparator();
    fileMenu->addAction(exitAct);

    scanMenu = menuBar()->addMenu(tr("&ScanMatcher"));
    scanMenu->addAction(useNoSMAct);
    scanMenu->addAction(useWeightedSMAct);
    scanMenu->addAction(useUnweightedSMAct);

    viewMenu = menuBar()->addMenu(tr("&View"));
    viewMenu->addAction(zoomInAct);
    viewMenu->addAction(zoomOutAct);
    viewMenu->addAction(zoomResetAct);


    windowMenu = menuBar()->addMenu(tr("&Window"));
    connect(windowMenu, SIGNAL(aboutToShow()), this, SLOT(updateWindowMenu()));

    menuBar()->addSeparator();

    helpMenu = menuBar()->addMenu(tr("&Help"));
    helpMenu->addAction(aboutAct);
}

void MainWindow::createToolBars()
{
    fileToolBar = addToolBar(tr("File"));
    fileToolBar->addAction(openAct);
    fileToolBar->addAction(saveAct);

    viewToolBar = addToolBar(tr("View"));
    viewToolBar->addAction(zoomInAct);
    viewToolBar->addAction(zoomOutAct);
    viewToolBar->addAction(zoomResetAct);

}

void MainWindow::createStatusBar()
{
    statusBar()->showMessage(tr("Ready"));
}

void MainWindow::readSettings()
{
    QSettings settings("ScanMatcher", "MainWindow");
    QPoint pos = settings.value("pos", QPoint(200, 200)).toPoint();
    QSize size = settings.value("size", QSize(400, 400)).toSize();
    move(pos);
    resize(size);
}

void MainWindow::writeSettings()
{
    QSettings settings("ScanMatcher", "MainWindow");
    settings.setValue("pos", pos());
    settings.setValue("size", size());
}

MapWindow *MainWindow::activeMapWindow()
{
    return qobject_cast<MapWindow *>(workspace->activeWindow());
}

MapWindow *MainWindow::findMapWindow(const QString &fileName)
{
    QString canonicalFilePath = QFileInfo(fileName).canonicalFilePath();

    foreach (QWidget *window, workspace->windowList()) {
        MapWindow *child = qobject_cast<MapWindow *>(window);
        if (child->currentFile() == canonicalFilePath)
            return child;
    }
    return 0;
}
